;; The first three lines of this file were inserted by DrRacket. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "htdp-beginner-reader.ss" "lang")((modname model-list-program) (read-case-sensitive #t) (teachpacks ((lib "image.ss" "teachpack" "2htdp") (lib "universe.ss" "teachpack" "2htdp"))) (htdp-settings #(#t constructor repeating-decimal #f #t none #f ((lib "image.ss" "teachpack" "2htdp") (lib "universe.ss" "teachpack" "2htdp")))))
#|
We are given a list of lists of Strings. Some of the Strings appear in 
several of the lists, some appear just once. Your task is to produce 
a list of all Strings that appear in the original list, but each String 
can only appear once in the resulting list.

Remember, we follow the design recipe!
|#

;;------------------------------------------------------------------------
;; THE MAIN FUNCTION

;; Step 1: Problem analysis and data definitions.
;; These were already given above.

;; Step 2: Purpose statement, contract, and function header.

;; simple-list: [Listof [Listof String]] -> [Listof String]
;; produce a list of Strings that occur in the given Listof Listof Strings
;; (define (simple-list lists) ...)

;; Step 3: Examples of data.

;; The following is an example of data we may consume:

(define alists (list (list "bear" "mosquito" "fish")
                     (list "horse" "bear")
                     (list "fish" "horse" "mouse")
                     (list "mouse" "bear")))

;; and we expect that the result will be a list that contains
;; the following Strings:

(define result (list "bear" "mosquito" "fish" "horse" "mouse"))

;; Step 4: Template

;(define (simple-list lists) 
;  (cond
;    [(empty? lists) ... ]
;    [(cons? lists)  ... (first lists) ...
;                    ... (rest lists) ...
;                    ... (simple-list (rest lists)) ...]))

;For the example above, we have the following pieces of data available
;in the 'cons?' case:
;
;(first lists)   -------------- (list "bear" "mosquito" "fish")
;(rest lists)    -------------- (list (list "horse" "bear")
;                                     ("fish" "horse" "mouse")
;                                     ("mouse" "bear")))
;(simple-list (rest lists)) --- (list "horse" "bear" "fish" "mouse")
;             
;Looking at the expected result we see that we need to add to the
;(simple-lists (rest lists)) the String "mosquito", but none of the
;others. The problem here involves two lists --- and so should 
;definitely be handled by a separate function.

;;------------------------------------------------------------------------
;; THE WISH LIST

;; Step 1: Problem analysis and data definition

;The new function consumes two lists, the source and the accumulator.
;It adds to the accumulator every element from the source that is not
;already contained in the accumulator.

;; Step 2: Purpose statement, contract, and function header.

;; add-to-list: [Listof String] [Listof String] -> [Listof String]
;; produce a new list by adding to the accumulator all Strings 
;; from the source that it does not already contain
;(define (add-to-list src acc) ...)

;; Step 3: Examples of data.

;; The following example has motivated this wish list function:

;src                   --- (list "bear" "mosquito" "fish")
;acc                   --- (list "horse" "bear" "fish" "mouse")
;(add-to-list src acc) --- (list "mosquito" "horse" "bear" "fish" "mouse")

;We will finish the design of this function later.
;We return to the design of the main function:

;;------------------------------------------------------------------------
;; THE MAIN FUNCTION

;; Step 4: Template

;(define (simple-list lists) 
;  (cond
;    [(empty? lists) ... ]
;    [(cons? lists)  
;        ... (first lists) ...                 [Listof String]
;        ... (rest lists) ...                  [Listof [Listof String]]
;        ... (simple-list (rest lists)) ...    [Listof String]
;        ... (add-to-list [Listof String] [Listof String])) ...]))

;; Step 5: Function Body

;; This follows now easily from the purpose statement for the add-to-list
;; function:

;(define (simple-list lists)
;  (cond
;    [(empty? lists) empty ]
;    [(cons? lists)  (add-to-list (first lists) 
;                                 (simple-list (rest lists)))]))
  
;; Step 6: Tests

;testing the method simple-list
;(equal? (simple-list empty) empty)
;(equal? (simple-list '(empty)) empty)
;(equal? (simple-list alists) result)

;; Of course, we cannot run the tests until we have finished designing
;; the wish list method

;;------------------------------------------------------------------------
;; THE WISH LIST

;; Step 4: Template

;(define (add-to-list src acc) 
;  (cond
;    [(empty? src) acc]
;    [(cons? src)  
;        ... (first src) ...                 [Listof String]
;        ... (rest src) ...                  [Listof String]
;        ... (add-to-list [Listof String] [Listof String]) ...]))

;; Step 5: Function Body

;; From the example we see that we only add (first src) to acc 
;; if acc does not already contain (first src). 
;; We need another helper function: 'contains?'
;; We leave it to you to design this function.
;; However, we add its application to the template:

;;           does acc contains the String (first src)?
;        ... (contains? acc (first src)) ...

;(define (add-to-list src acc) 
;  (cond
;    [(empty? src) acc]
;    [(cons? src)  
;        (cond 
;          [(contains? acc (first src)) (add-to-list (rest src) acc)]
;          [else (add-to-list (rest src) acc)])]))

;; Step 6: Tests

;testing the method add-to-list
;(equal? (add-to-list empty empty) empty)
;(equal? (add-to-list empty (list "hello")) (list "hello"))
;(equal? (add-to-list (list "hello") empty) (list "hello"))
;(equal? (add-to-list (list "hi" "bye" "hello") 
;        (list "hello")) (list "hi" "bye" "hello") )
;(equal? (add-to-list (list "hi" "ciao") (list "hello")) 
;        (list "hi" "ciao" "hello") )
;(equal? (add-to-list (list "hello" "ciao")  (list "hi" "bye" "hello")) 
;        (list "hello"))

;;------------------------------------------------------------------------
;; PUTTING IT ALL TOGETHER

(define (contains? alist str)
  (cond
    [(empty? alist) false]
    [else (or (equal? (first alist) str)
              (contains? (rest alist) str))]))
  
  (define (add-to-list src acc) 
  (cond
    [(empty? src) acc]
    [(cons? src)  
        (cond 
          [(contains? acc (first src)) (add-to-list (rest src) acc)]
          [else (cons (first src) (add-to-list (rest src) acc))])]))

;testing the method add-to-list
(equal? (add-to-list empty empty) empty)
(equal? (add-to-list empty (list "hello")) (list "hello"))
(equal? (add-to-list (list "hello") empty) (list "hello"))
(equal? (add-to-list (list "hi" "bye" "hello") (list "hello")) 
        (list "hi" "bye" "hello"))
(equal? (add-to-list (list "hi" "ciao") (list "hello")) 
        (list "hi" "ciao" "hello"))
(equal? (add-to-list (list "hello" "ciao")  (list "hi" "bye" "hello")) 
        (list "ciao" "hi" "bye" "hello"))

(define (simple-list lists)
  (cond
    [(empty? lists) empty ]
    [(cons? lists)  
     (add-to-list (first lists) (simple-list (rest lists)))]))

(equal? (simple-list empty) empty)
(equal? (simple-list (list empty)) empty)
(equal? (simple-list alists) result)

;; Note that the last test fails. We did not require any special ordering 
;; of the resulting list. Indeed, looking at
(simple-list alists)

;; we see that it produces 
;; (list "mosquito" "fish" "horse" "mouse" "bear")

;; a list that contains the same Strings as our expected result:
;; (list "bear" "mosquito" "fish" "horse" "mouse")

;; To really finish the job, we would need to design a test for just this
;; property of the result, namely, that the second list contains 
;; every String that appears in the first list, no other Strings, and 
;; no String appears there more than once.

;; NOTE: For actual tests we should use the 'check-expect' test evaluation 
;;       We have used a simple comparison, so that we can highlight the
;;       problem with the last test.

